import asyncio
import gzip
import json
import logging
import time
from enum import Enum
from random import random
from typing import Optional, List, Dict, Callable, Any

import websockets as ws
from ftx import AsyncClient
from ftx import ws_signature
from .threaded_stream import ThreadedApiManager

KEEPALIVE_TIMEOUT = 5 * 60  # 5 minutes


class WSListenerState(Enum):
    INITIALISING = "Initialising"
    STREAMING = "Streaming"
    RECONNECTING = "Reconnecting"
    EXITING = "Exiting"


class ReconnectingWebsocket:
    MAX_RECONNECTS = 5
    MAX_RECONNECT_SECONDS = 60
    TIMEOUT = 60

    def __init__(
        self,
        loop,
        url: str,
        name: Optional[str] = None,
        is_binary: bool = False,
        exit_coro=None,
    ):
        self._loop = loop or asyncio.get_event_loop()
        self._log = logging.getLogger(__name__)
        self._name = name
        self._url = url
        self._exit_coro = exit_coro
        self._reconnects = 0
        self._is_binary = is_binary
        self._conn = None
        self._socket = None
        self.ws: Optional[ws.WebSocketClientProtocol] = None
        self.ws_state = WSListenerState.INITIALISING
        self._queue = asyncio.Queue(loop=self._loop)

    async def __aenter__(self):
        await self.connect()
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._exit_coro:
            await self._exit_coro(self._name)
        self.ws_state = WSListenerState.EXITING
        if self.ws:
            self.ws.fail_connection()
        if self._conn:
            await self._conn.__aexit__(exc_type, exc_val, exc_tb)
        self.ws = None

    async def connect(self):
        await self._before_connect()
        assert self._name
        self._conn = ws.connect(self._url, close_timeout=0.001)
        try:
            self.ws = await self._conn.__aenter__()
        except:  # noqa
            print("reconnect")
            await self._reconnect()
            return
        self.ws_state = WSListenerState.STREAMING

        self._reconnects = 0
        await self._after_connect()
        self._loop.call_soon_threadsafe(asyncio.create_task, self._read_loop())

    async def _run(self):
        pass

    async def send_msg(self, msg):
        while not self.ws:
            time.sleep(0.1)
        await self.ws.send(json.dumps(msg))

    async def _before_connect(self):
        pass

    async def _after_connect(self):
        pass

    def _handle_message(self, evt):
        if self._is_binary:
            try:
                evt = gzip.decompress(evt)
            except (ValueError, OSError):
                return None
        try:
            return json.loads(evt)
        except ValueError:
            self._log.debug(f"error parsing evt json:{evt}")
            return None

    async def _read_loop(self):
        while True:
            res = None
            if not self.ws or self.ws_state != WSListenerState.STREAMING:
                await self._wait_for_reconnect()
                break
            if self.ws_state == WSListenerState.EXITING:
                break
            if self.ws.state == ws.protocol.State.CLOSING:
                break
            if self.ws.state == ws.protocol.State.CLOSED:
                try:
                    print("close")
                    await self._reconnect()
                except Exception as e:
                    print(e)
                else:
                    break
            try:
                res = await asyncio.wait_for(self.ws.recv(), timeout=self.TIMEOUT)
            except asyncio.TimeoutError:
                logging.debug(f"no message in {self.TIMEOUT} seconds")
                print("timeout")
                await self._reconnect()
            except asyncio.CancelledError as e:
                logging.debug(f"cancelled error {e}")
                break
            except asyncio.IncompleteReadError as e:
                logging.debug(f"incomplete read error {e}")
            except Exception as e:
                logging.debug(f"exception {e}")
                await self._reconnect()
                break
            else:
                if self.ws_state in (
                    WSListenerState.EXITING,
                    WSListenerState.RECONNECTING,
                ):
                    break
                res = self._handle_message(res)
                if self.ws_state in (
                    WSListenerState.EXITING,
                    WSListenerState.RECONNECTING,
                ):
                    break

            if res and self._queue.qsize() < 100:
                await self._queue.put(res)

    async def recv(self):
        res = None
        while not res:
            try:
                res = await asyncio.wait_for(self._queue.get(), timeout=self.TIMEOUT)
            except asyncio.TimeoutError:
                logging.debug(f"no message in {self.TIMEOUT} seconds")
        return res

    async def _wait_for_reconnect(self):
        while self.ws_state == WSListenerState.RECONNECTING:
            logging.debug("reconnecting waiting for connect")
        if not self.ws:
            logging.debug("ignore message no ws")
        else:
            logging.debug(f"ignore message {self.ws_state}")

    def _get_reconnect_wait(self, attempts: int) -> int:
        expo = 2 ** attempts
        return round(random() * min(self.MAX_RECONNECT_SECONDS, expo - 1) + 1)

    async def before_reconnect(self):
        if self.ws:
            await self._conn.__aexit__(None, None, None)
            self.ws = None
        self._reconnects += 1

    def _no_message_received_reconnect(self):
        logging.debug("No message received, reconnecting")
        asyncio.create_task(self._reconnect())

    async def _reconnect(self):
        if self.ws_state == WSListenerState.RECONNECTING:
            return
        self.ws_state = WSListenerState.RECONNECTING
        await self.before_reconnect()
        if self._reconnects < self.MAX_RECONNECTS:
            reconnect_wait = self._get_reconnect_wait(self._reconnects)
            logging.debug(
                f"websocket reconnecting {self.MAX_RECONNECTS - self._reconnects} reconnects left - "
                f"waiting {reconnect_wait}"
            )
            await asyncio.sleep(reconnect_wait)
            await self.connect()
        else:
            logging.error(f"Max reconnections {self.MAX_RECONNECTS} reached:")
            raise "MaximumReconnectRetry"


class FtxSocketManager:
    WS_URL = "wss://ftx.com/ws/"

    def __init__(
        self,
        client: AsyncClient,
        loop=None,
    ):
        self._conns = {}
        self._loop = loop or asyncio.get_event_loop()
        self._client = client
        self._log = logging.getLogger("FtxSocketManager")

    def _init_stream_url(self):
        self.ws_url = self.STREAM_URL
        self.private_ws_url = self.PSTREAM_URL

    def _get_socket(self, socket_name: str, is_binary: bool = False) -> str:
        conn_id = f"{socket_name}"
        if conn_id not in self._conns:
            self._conns[conn_id] = ReconnectingWebsocket(
                loop=self._loop,
                name=socket_name,
                url=self.WS_URL,
                exit_coro=self._exit_socket,
                is_binary=is_binary,
            )

        return self._conns[conn_id]

    async def subscribe(self, conn_name: str, **params):
        try:

            await self._conns[conn_name].send_msg(params)
        except KeyError:
            self._log.warning(f"Connection name: <{conn_name}> not create and start!")

    async def _exit_socket(self, name: str):
        await self._stop_socket(name)

    def get_socket(self, socket_name):
        return self._get_socket(socket_name)

    async def _stop_socket(self, conn_key):
        if conn_key not in self._conns:
            return

        del self._conns[conn_key]


class ThreadedWebsocketManager(ThreadedApiManager):
    def __init__(
        self,
        api_key: Optional[str] = None,
        api_secret: Optional[str] = None,
        subaccount: str = None,
    ):
        super().__init__(api_key, api_secret)
        self._bsm: Optional[FtxSocketManager] = None
        self.api = api_key
        self.secret = api_secret
        self.subaccount = subaccount

    async def _before_socket_listener_start(self):
        assert self._client
        self._bsm = FtxSocketManager(client=self._client, loop=self._loop)

    def _start_socket(
        self,
        callback: Callable,
        socket_name: str,
    ) -> str:
        while not self._bsm:
            time.sleep(0.1)

        socket = getattr(self._bsm, "get_socket")(socket_name)
        name = socket._name
        self._socket_running[name] = True
        self._loop.call_soon_threadsafe(
            asyncio.create_task,
            self.start_listener(socket, socket._name, callback, self.ping),
        )

        return socket

    def start_socket(
        self,
        callback: Callable,
        socket_name: str,
    ) -> str:
        socket = self._start_socket(
            callback=callback,
            socket_name=socket_name,
        )
        self.ping(socket_name)
        return socket

    def subscribe(self, conn_name: str, **params):
        while not self._bsm:
            time.sleep(0.1)
        try:
            loop = asyncio.get_running_loop()
        except RuntimeError:
            loop = None
        if loop and loop.is_running():
            loop.create_task(self._bsm.subscribe(conn_name, **params))
        else:
            asyncio.run(self._bsm.subscribe(conn_name, **params))

    def login(self, conn_name: str):
        if not conn_name:
            return
        ts = int(time.time() * 1000)
        sign = ws_signature(ts, self.secret)
        args = {}
        args["key"] = self.api
        args["sign"] = sign
        args["time"] = ts
        if self.subaccount != None:
            args["subaccount"] = self.subaccount

        self.subscribe(conn_name=conn_name, args=args, op="login")

    def ping(self, name):
        self.subscribe(name, op="ping")



import json


class FtxAPIException(Exception):
    def __init__(self, resp_json, status_code):
        self.message = resp_json["message"]
        self.code = status_code
        self.api_code = resp_json["code"]

    def __str__(self):
        return f"APIError(code={self.api_code}): {self.message}"


class FtxValueError:
    def __init__(self, response) -> None:
        self.response = response

    def __str__(self):
        return f"Invalid Response: {self.response.text}"



import time
import json
import aiohttp
import asyncio
import requests
from ftx import signature
from urllib.parse import urlencode
from typing import Dict, Optional, List, Tuple
from ftx import FtxAPIException, FtxValueError
from concurrent.futures._base import TimeoutError


class BaseClient:
    API_URL = "https://ftx.com/api"

    def __init__(self, api: Optional[str] = None, secret: Optional[str] = None):
        self.API_KEY, self.API_SECRET = api, secret
        self.session = self._init_session()
        self.header = {}
        self.TIMEOUT = 45

    def _get_header(self) -> Dict:
        header = {
            "FTX-KEY": "",
            "FTX-SIGN": "",
            "FTX-TS": "",
        }
        if self.API_KEY:
            assert self.API_KEY
            header["FTX-KEY"] = self.API_KEY
        return header

    def _init_session(self):
        raise NotImplementedError

    def _init_url(self):
        self.api_url = self.API_URL

    def _handle_response(self, response: requests.Response):
        code = response.status_code
        if code == 200:
            return response.json()
        else:
            try:
                resp_json = response.json()
                raise FtxAPIException(resp_json, code)
            except ValueError:
                raise FtxValueError(response)


class Client(BaseClient):
    def __init__(self, api: Optional[str], secret: Optional[str], subaccount=None):
        super().__init__(api=api, secret=secret)
        self.subaccount = subaccount

    def _init_session(self) -> requests.Session:
        self.header = self._get_header()
        session = requests.session()
        session.headers.update(self.header)
        return session

    def _get(self, path: str, params=None):
        return self._request("get", path, params)

    def _post(self, path: str, params=None) -> Dict:
        return self._request("post", path, params)

    def _put(self, path: str, params=None) -> Dict:
        return self._request("put", path, params)

    def _delete(self, path: str, params=None) -> Dict:
        return self._request("delete", path, params)

    def _request(self, method, path: str, params: Dict):
        try:
            ts = str(int(time.time() * 1000))
            uri = f"{self.API_URL}{path}"
            if method == "post":
                sig = signature(ts, method, path, self.API_SECRET, params)
            else:
                sig = signature(ts, method, path, self.API_SECRET)
            self.header["FTX-KEY"] = self.API_KEY
            self.header["FTX-SIGN"] = sig
            self.header["FTX-TS"] = ts
            if self.subaccount is not None:
                self.header["FTX-SUBACCOUNT"] = self.subaccount
            self.session.headers.update(self.header)
            self.response = getattr(self.session, method)(uri, json=params)
            return self._handle_response(self.response)
        except Exception as e:
            print(f"[ERROR] Request failed!")
            print(self.response.text)

    def get_markets(self) -> Dict:
        return self._get("/markets")

    def get_account_info(self) -> Dict:
        return self._get("/account")

    def get_all_balances(self) -> Dict:
        return self._get("/wallet/all_balances")

    def get_balances(self) -> Dict:
        return self._get("/wallet/balances")

    def get_positions(self) -> Dict:
        return self._get("/positions")

    def send_order(self, **kwargs) -> Dict:
        try:
            return self._post("/orders", params=kwargs)
        except:
            time.sleep(0.2)
            return self._post("/orders", params=kwargs)

    def set_leverage(self, **kwargs):
        return self._post("/account/leverage", kwargs)

    def request_withdrawal(self, **kwargs):
        return self._post("/wallet/withdrawals", kwargs)

    def cancel_order(self, order_id):
        return self._delete(f"/orders/{order_id}")


class AsyncClient(BaseClient):
    def __init__(
        self,
        api: Optional[str],
        secret: Optional[str],
        subaccount=None,
        loop=None,
    ):
        self.loop = loop or asyncio.get_event_loop()
        self.subaccount = subaccount
        super().__init__(
            api=api,
            secret=secret,
        )

    @classmethod
    async def create(
        cls,
        api: Optional[str],
        secret: Optional[str],
        loop=None,
    ):
        self = cls(api, secret, loop)
        return self

    def _init_session(self) -> aiohttp.ClientSession:
        session = aiohttp.ClientSession(loop=self.loop, headers=self._get_header())
        return session

    async def close_connection(self):
        if self.session:
            assert self.session
            await self.session.close()

    async def _request(self, method, uri: str, path: str, params: Dict):
        try:
            ts = str(int(time.time() * 1000))
            uri = f"{self.API_URL}{path}"

            sig = signature(ts, method, path, self.API_SECRET)
            self.header["FTX-KEY"] = self.API_KEY
            self.header["FTX-SIGN"] = sig
            self.header["FTX-TS"] = ts
            if self.subaccount is not None:
                self.header["FTX-SUBACCOUNT"] = self.subaccount
            self.session.headers.update(self.header)

            async with getattr(self.session, method)(uri, params=params) as response:
                self.response = response
                return await self._handle_response(response)

        except Exception as e:
            print(f"[ERROR] Request failed!")
            print(e)

    async def _handle_response(self, response: requests.Response):
        code = response.status_code
        if code == 200:
            return response.json()
        else:
            try:
                resp_json = response.json()
                raise FtxAPIException(resp_json, code)
            except ValueError:
                raise FtxValueError(response)

    async def _get(self, path: str, params=None):
        return await self._request("get", path, params)

    async def _post(self, path: str, params=None) -> Dict:
        return await self._request("post", path, params)

    async def _put(self, path: str, params=None) -> Dict:
        return await self._request("put", path, params)

    async def _delete(self, path: str, params=None) -> Dict:
        return await self._request("delete", path, params)

    async def get_available_symbol(self):
        return await self._get("public/info")



from ftx.timer import PerpetualTimer
from ftx.exceptions import FtxValueError, FtxAPIException
from ftx.authentication import signature, ws_signature
from ftx.client import Client, AsyncClient
from ftx.streams import ThreadedWebsocketManager



import asyncio
import threading
from ftx import PerpetualTimer
from typing import Callable, Optional, Dict
from ftx import AsyncClient


class ThreadedApiManager(threading.Thread):
    def __init__(
        self,
        api_key: Optional[str] = None,
        api_secret: Optional[str] = None,
    ):
        """Initialise the FtxSocketManager"""
        super().__init__()
        self._loop: asyncio.AbstractEventLoop = asyncio.new_event_loop()
        self._client: Optional[AsyncClient] = None
        self._running: bool = True
        self._socket_running: Dict[str, bool] = {}
        self._client_params = {
            "api": api_key,
            "secret": api_secret,
        }

    async def _before_socket_listener_start(self):
        ...

    async def socket_listener(self):
        self._client = await AsyncClient.create(loop=self._loop, **self._client_params)
        await self._before_socket_listener_start()
        while self._running:
            await asyncio.sleep(0.2)
        while self._socket_running:
            await asyncio.sleep(0.2)

    async def start_listener(self, socket, name: str, callback, ping: Optional[Callable] = None):
        async with socket as s:
            while self._socket_running[name]:
                try:
                    msg = await asyncio.wait_for(s.recv(), 3)
                except asyncio.TimeoutError:
                    ...
                    continue
                if not msg:
                    continue
                if "type" in msg and msg["type"] == "pong":
                    ping(name)
                    continue
                callback(msg)
        del self._socket_running[name]

    def run(self):
        self._loop.run_until_complete(self.socket_listener())

    def stop_socket(self, socket_name):
        if socket_name in self._socket_running:
            self._socket_running[socket_name] = False

    async def stop_client(self):
        await self._client.close_connection()

    def stop(self):
        if not self._running:
            return
        self._running = False
        self._loop.call_soon(asyncio.create_task, self.stop_client())
        for socket_name in self._socket_running.keys():
            self._socket_running[socket_name] = False



import hmac
import json


def signature(ts: str, method: str, path_url: str, secret: str, post_body=None):
    signature_payload = f"{ts}{method.upper()}/api{path_url}".encode()
    if method == "post":
        signature_payload += json.dumps(post_body).encode()
    signature = hmac.new(secret.encode(), signature_payload, "sha256").hexdigest()
    return signature


def ws_signature(ts: str, secret: str):
    signature_payload = f"{ts}websocket_login".encode()
    signature = hmac.new(secret.encode(), signature_payload, "sha256").hexdigest()
    return signature



SYMBOL_TYPE_SPOT = "SPOT"

ORDER_STATUS_NEW = "NEW"
ORDER_STATUS_PARTIALLY_FILLED = "PARTIAL_FILLED"
ORDER_STATUS_FILLED = "FILLED"
ORDER_STATUS_CANCELED = "CANCELLED"
ORDER_STATUS_REJECTED = "REJECTED"
ORDER_STATUS_INCOMPLETE = "INCOMPLETE"
ORDER_STATUS_COMPLETED = "COMPLETED"

KLINE_INTERVAL_1MINUTE = "1m"
KLINE_INTERVAL_5MINUTE = "5m"
KLINE_INTERVAL_15MINUTE = "15m"
KLINE_INTERVAL_30MINUTE = "30m"
KLINE_INTERVAL_1HOUR = "1h"
KLINE_INTERVAL_1DAY = "1d"
KLINE_INTERVAL_1WEEK = "1w"
KLINE_INTERVAL_1MONTH = "1M"

SIDE_BUY = "BUY"
SIDE_SELL = "SELL"

ORDER_TYPE_LIMIT = "LIMIT"
ORDER_TYPE_MARKET = "MARKET"
ORDER_TYPE_IOC = "IOC"
ORDER_TYPE_FOK = "FOK"
ORDER_TYPE_POST_ONLY = "POST_ONLY"
ORDER_TYPE_ASK = "ASK"
ORDER_TYPE_BID = "BID"



from threading import Timer


class PerpetualTimer:
    def __init__(self, t, hFunction):
        self.t = t
        self.hFunction = hFunction
        self.thread = Timer(self.t, self.handle_function)

    def handle_function(self):
        self.hFunction()
        self.thread = Timer(self.t, self.handle_function)
        self.thread.start()

    def start(self):
        self.thread.start()

    def cancel(self):
        self.thread.cancel()



from __future__ import print_function
from setuptools import setup, find_packages
import sys

setup(
    name="python-ftx",
    version="0.0.4",
    author="tinghsuwan",
    author_email="wanth1997@gmail.com",
    description="FTX python API SDK",
    license="MIT",
    url="https://github.com/wanth1997/python-ftx",
    packages=["ftx"],
    include_package_data=True,
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    install_requires=["requests", "aiohttp", "websockets"],
    zip_safe=True,
)
